import logging
import sys
import time

# from logging import debug, info, warning, error, critical
from typing import Callable, Dict, Optional

from colorama import Back, Fore, Style
from stratosphere import config


class ColoredFormatter(logging.Formatter):
    """
    Handles the formatting of coloured logging messages.
    """

    def __init__(self, *args, colors: Optional[Dict[str, str]] = None, **kwargs) -> None:
        """Create a new color formatter.

        Args:
            colors (Optional[Dict[str, str]], optional): . Dictionary of colors and logging levels. Defaults to None.
            args: args to pass to logging.Formatter.
            kwargs: kwargs to pass to logging.Formatter.
        """
        super().__init__(*args, **kwargs)
        self.colors = colors if colors else {}

    def format(self, record) -> str:  # noqa
        """Format the record as coloured text

        Args:
            record (_type_): logging record to format.

        Returns:
            str: Resulting formatted coloured logging message.
        """

        record.color = self.colors.get(record.levelname, "")
        record.reset = Style.RESET_ALL
        return super().format(record)


def init_logging(level: int):
    """Logging intialization.

    Args:
        level (int): logging level to activate.
    """

    formatter = ColoredFormatter(
        "{color}■{reset} {message}",
        style="{",
        datefmt="%Y-%m-%d %H:%M:%S",
        colors={
            "DEBUG": Fore.CYAN,
            "INFO": Fore.GREEN,
            "WARNING": Fore.YELLOW,
            "ERROR": Fore.RED,
            "CRITICAL": Fore.RED + Back.WHITE + Style.BRIGHT,
        },
    )

    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(formatter)

    logger = logging.getLogger()
    logger.handlers[:] = []
    logger.addHandler(handler)
    logger.setLevel(level)

    logging.info("Logging initialized.")


def timeit(f: Callable) -> Callable:
    """Decorator to measure the call execution time.

    Args:
        f (Callable): Function to execute

    Returns:
        Callable: Function with decorator.
    """

    def timed(*args, **kw):
        ts = time.perf_counter()
        result = f(*args, **kw)
        te = time.perf_counter()
        logging.info(f"Elapsed time @{f.__name__}: {(te - ts):.2f}s")
        return result

    return timed


class IgnoredChainedCallLogger:
    """
    This class is designed to catch chained operations, logging that they won't
    be executed, and returning nicely."""

    def __getitem__(self, index):
        """Catching all field requests, logging them.

        Args:
            index (_type_): index to be resolved.

        Returns:
            IgnoredChainedCallLogger: Recursive instance of the same class.
        """
        logging.error(f"Ignoring chained operation: .[{index}]")
        return IgnoredChainedCallLogger()

    def __getattribute__(self, name):
        """Catching all attribute requests, logging them.

        Args:
            name (_type_): name to be resolved.

        Returns:
            IgnoredChainedCallLogger: Recursive instance of the same class.
        """

        if name == "_ipython_canary_method_should_not_exist_":
            # this is handled separately, to handle how ipython is
            # managing the objects at the end of cells.
            raise AttributeError()

        if name in ["_repr_html_", "__str__"]:
            # empty output.

            return lambda: ""

        if name not in ["__class__", "_ipython_display_"]:
            # in case a new chained operation is requested, log it.
            logging.error(f"Ignoring chained operation: .{name}(...)")

        def newfunc(*args, **kwargs):
            # recursive instantiation of IgnoredChainedCallLogger
            return IgnoredChainedCallLogger()

        return newfunc

    @staticmethod
    def _repr_html_() -> str:
        """HTML representation, empty: we already produce the logging output.

        Returns:
            str: Empty string.
        """
        return ""


def fatal(*args, **kwargs):
    """Handle fatal situations, logging them and handling over the execution
    to IgnoredChainedCallLogger to log chained operations.

    Returns:
        IgnoredChainedCallLogger: Logger of chained opeartions.
    """

    logging.error(*args, **kwargs)
    return IgnoredChainedCallLogger()


def default_exception_handler(func: Callable) -> Callable:
    """Decorator Handle exceptions occurring in chained function calls.
    e.g., (a().b().c())
    It does not always work with subscripts, e.g., x[123].

    Args:
        func (Callable): Function to decorate.

    Returns:
        Callable: Decorated function.
    """

    def inner_function(*args, **kwargs):
        if not config.log.catch_exceptions:
            return func(*args, **kwargs)
        else:
            try:
                return func(*args, **kwargs)
            except (KeyboardInterrupt, TypeError, Exception) as e:
                if type(e) == KeyboardInterrupt:
                    return fatal("Keyboard interrupt.")
                else:
                    return fatal(f"Exception {e.__class__.__name__} at .{func.__name__}(...): {e}")

    return inner_function



from typing import Callable, List, Union

from stratosphere.store.db.database import next_ulid
import pandas as pd

from stratosphere.store.frames import reorder_columns


class RunNotFoundException(Exception):
    """Exception raised if the run is not found."""

    def __init__(self, id_run):
        self.message = f"Run '{id_run}' not found. List with .df() the available runs."
        super().__init__(self.message)


class Runs(dict):
    def __init__(self, runs=None):
        if runs is None:
            runs = []

        if isinstance(runs, Runs):
            runs = runs.values()

        super().__init__({str(run.id_run): run for run in runs})

    def get_keys(self, name):
        if len(self) == 0:
            return []
        else:
            first_key = next(iter(self))
            return list(getattr(self.get(first_key), name).keys())

    def df(self, max_level=None):
        # Load the dataframe from a dict or list of dicts.

        if len(self) == 0:
            # no runs!
            return pd.DataFrame(columns=["id_run"])

        df = pd.json_normalize([{**run.fields, **{"id_run": run.id_run}} for run in self.values()], max_level=max_level)

        return reorder_columns(df, ["id_run"])


class Run:
    """A run represents an instance of the experiment, obtained by
    combining the fixed and variable parameters. The Run objects
    (and the tracked fields)  must be serializable with cloudpickle.
    """

    def __init__(
        self,
        id_run: str = None,
        funcs: Union[Callable, List[Callable]] = None,
        kwargs: dict = None,
        params: dict = None,
        fields: dict = None,
    ):
        """Create a new run.

        Args:
            kwargs (dict, optional): Fixed parameters for all runs of an experiment. Defaults to None.
            parameters (dict, optional): Variable parameters to be considered for this run. Defaults to None.
            fields (dict, optional): fields to be tracked. Defaults to None.
            funcs (Union[Callable, List[Callable]], optional): One or more functions to be executed. Their
            only parameter is an instance of the Run class itself.
            experiment_name (str, optional): Name of the experiment the run belongs to. Defaults to None.
        """

        self.id_run = next_ulid() if id_run is None else id_run
        self.funcs = normalize_funcs(funcs)
        self.kwargs = {} if kwargs is None else kwargs
        self.params = {} if params is None else params
        self.fields = {} if fields is None else fields
        self.exception = None

    def execute(self, funcs=None):
        if funcs is not None:
            self.funcs = normalize_funcs(funcs)

        self.exception = None

        for func in self.funcs:
            try:
                func(self)
            except Exception as e:  # noqa
                # We want to catch all exceptions in the distributed execution of runs,
                # It will then be reported by Experiment.execute.
                self.exception = f"{func.__qualname__.split('.')[0]}: {str(e)}"
                break

        return self

    def verify(self, func):
        func(self)


def normalize_funcs(funcs):
    if funcs is None:
        return []
    elif callable(funcs):
        return [funcs]
    else:
        return funcs



from stratosphere.run import Run


def load_datasets(datasets):
    def func(run: Run):
        for key, key_type in key_types:
            assert_keys(name, key)
            if not isinstance(getattr(run, name)[key], key_type):
                raise RunAssertion(f"type(run.{name}) != {key_type}")

    return func



from typing import List
import pandas as pd


def reorder_columns(df: pd.DataFrame, ordered_columns: List[str]) -> List[str]:
    """Given a dataframe, enforce a partial ordering of some columns, appending
    the remaining ones in sorted order.

    Args:
        df (pd.DataFrame): Consider the columns from thsi dataframe.
        ordered_columns (List[str]): Consider this partial ordering of the columns.

    Returns:
        List[str]: New ordering of all columns for the {df} dataframe.
    """

    remaining_columns = sorted([col_name for col_name in df.columns if col_name not in ordered_columns])
    return df[ordered_columns + remaining_columns]



import pandas as pd
from stratosphere.run import Run


class RunAssertion(Exception):
    """Raise this exception if an assertion on the runs state fails."""

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


def assert_keys(name, keys, missing=False):
    if isinstance(keys, str):
        keys = [keys]

    def func(run: Run):
        for key in keys:
            if not missing and (key not in getattr(run, name)):
                raise RunAssertion(f"run.{name}['{key}'] missing.")
            if missing and (key in run[name]):
                raise RunAssertion(f"run.{name}['{key}'] already present.")

    return func


def assert_types(name, key_types):
    def func(run: Run):
        for key, key_type in key_types:
            assert_keys(name, key)
            if not isinstance(getattr(run, name)[key], key_type):
                raise RunAssertion(f"type(run.{name}) != {key_type}")

    return func


def assert_df(name, key, columns=None, n_rows=None):
    def func(run: Run):
        assert_types(name, {key: pd.DataFrame})
        if columns is not None:
            cols_df = list(getattr(run, name)[key].columns)
            cols_expected = list(columns)
            if set(cols_df) != set(cols_expected):
                raise RunAssertion(f"run.{name}['{key}'].columns == {cols_df} != {cols_expected}")
        if n_rows is not None:
            n_rows_df = len(getattr(run, name)[key])
            if n_rows_df != n_rows:
                raise RunAssertion(f"len(run.{name}['{key}']) == {n_rows_df} != {n_rows}")

    return func



from functools import partial
import json
from typing import List
import uuid
import zlib
from pickle import PickleError

import cloudpickle
import numpy as np
import pandas as pd
from stratosphere import config


def compress(data: bytes, enable_compression=config.serialization.enable_compression) -> bytes:
    if enable_compression:
        return zlib.compress(data)
    else:
        return data


def decompress(data: bytes) -> bytes:
    if isinstance(data, memoryview):
        data = data.tobytes()

    try:
        data = zlib.decompress(data)
    except zlib.error:
        pass

    return data


# Introduced with Python 3.0, https://peps.python.org/pep-3154/
# Unpickling might fail: on different architectures, different python version, in case of missing packages.
PICKLE_DEFAULT_PROTOCOL = 4


def pickle_dumps(obj: object) -> bytes:
    """It returns the compressed (zlib) pickled object. The Pickle DEFAULT_PROTOCOL is used
    for maximum compatibility.

    Args:
        obj (object): Object to serialize.

    Returns:
        bytes: Compressed serialized object.
    """
    return compress(cloudpickle.dumps(obj, protocol=PICKLE_DEFAULT_PROTOCOL))


def pickle_loads(data: bytes) -> object:
    """It returns the loaded object, afer uncompressing and unpickling it.

    Args:
        data (bytes): Serialized object.

    Returns:
        object: Unserialized object.
    """
    try:
        return cloudpickle.loads(decompress(data))
    except Exception:  # noqa # we do want to catch all errors here
        raise PickleError()


def pickle_size(obj: object, unit: str = "b") -> int:
    """It returns the size of the object once serialised (including compression).

    Args:
        obj (object): Object to analyse.
        unit (str, optional): Unit of measure, b (Bytes), kb (KiloBytes), mb (MegaBytes). Defaults to "b".

    Returns:
        int: Size of the serialized object.
    """
    size_object = len(pickle_dumps(obj))
    if unit == "b":
        return int(size_object * 1e2) / 1e2
    elif unit == "kb":
        return int(size_object * 1e-3 * 1e2) / 1e2
    elif unit == "mb":
        return int(size_object * 1e-6 * 1e2) / 1e2
    else:
        return None


serialized_types = [pd.DataFrame, pd.Series, np.ndarray, uuid.UUID, dict, list]


class JSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, pd.DataFrame):
            return {"__type": "pandas.DataFrame", "data": obj.to_dict(orient="list"), "index": obj.index.tolist()}
        elif isinstance(obj, pd.Series):
            return {"__type": "pandas.Series", "data": obj.tolist(), "index": obj.index.tolist()}
        elif isinstance(obj, np.ndarray):
            return {"__type": "numpy.ndarray", "data": obj.tolist()}
        elif isinstance(obj, uuid.UUID):
            return {"__type": "uuid.UUID", "data": str(obj)}
        else:
            return json.JSONEncoder.default(self, obj)


def serialize(obj: object, enable_compression=config.serialization.enable_compression) -> bytes:
    return compress(json.dumps(obj, cls=JSONEncoder).encode("UTF-8"), enable_compression=enable_compression)


def deserialize(obj: bytes) -> object:
    def f(d):
        if isinstance(d, dict) and "__type" in d:
            # This ia a value to deserialize
            if d["__type"] == "pandas.DataFrame":
                df = pd.DataFrame.from_dict(d["data"], orient="columns")
                df.index = pd.Index(d["index"])
                return df
            elif d["__type"] == "pandas.Series":
                return pd.Series(d["data"], index=d["index"])
            elif d["__type"] == "numpy.ndarray":
                return np.asarray(d["data"])
            elif d["__type"] == "uuid.UUID":
                return uuid.UUID(d["data"])
            else:
                # We don't know how to deserialize it, return it as a dictionary.
                return d

        if isinstance(d, list):
            # Walk thru the list, trying to decode values.
            return [f(v) for v in d]

        if isinstance(d, dict):
            # Walk thru the dict, trying to decode values.
            return {kv[0]: f(kv[1]) for kv in d.items()}

        else:
            # Nothing to do, return value.
            return d

    return json.loads(decompress(obj).decode("UTF-8"), object_hook=f)


def serialize_df(
    df: pd.DataFrame, ignore_columns: list = None, enable_compression=config.serialization.enable_compression
):
    consider_columns = [col_name for col_name in df.columns if col_name not in ignore_columns]

    # Identify columns to serialize
    serialized_cols = []
    for col_name in consider_columns:
        for serialized_type in serialized_types:
            # We assume that the types in the first row are the same of the other rows.
            if isinstance(df[col_name].iloc[0], serialized_type):
                serialized_cols.append(col_name)
                break

    if len(serialized_cols) > 0:
        # If there are columns to serialize, work on a frame copy.
        df = df.copy()
        for col_name in serialized_cols:
            df[col_name] = df[col_name].map(partial(serialize, enable_compression=enable_compression))

    # Identify columns that haven't been serialized, among the ones to be considered.
    non_serialized_cols = [col_name for col_name in consider_columns if col_name not in serialized_cols]

    columns = {"serialized": serialized_cols, "non_serialized": non_serialized_cols, "compression": enable_compression}
    return df, columns


def deserialize_df(df: pd.DataFrame, columns: List[str]):
    if len(columns) == 0:
        return df

    df = df.copy()
    for col_name in columns:
        df[col_name] = df[col_name].map(deserialize)

    return df


def explode_json_column(df: pd.DataFrame, col_name: str, prefix: str = None, suffix: str = None) -> pd.DataFrame:
    """Explode a column in the Pandas dataframe containing a dict to a list of columns.
    Useful to handle the "attributes" column in the "Experiments" table, which contains
    the serialized values as JSON.
    Args:
        df (pd.DataFrame): Pandas dataframe
        col_name (str): Column to process.
        prefix (str, optional): Prefix to consider for all exploded columns.
        suffix (str, optional): In case exploded columns already exist, use this suffix.
        Defaults to _{col_name}.
    Returns:
        pd.DataFrame: Resulting Pandas dataframe.
    """

    if suffix is None:
        suffix = f"_{col_name}"

    # Explode column containing json to multiple columns, in a dataframe.
    df_exploded = pd.json_normalize(df[col_name])

    if prefix is not None:
        df_exploded.columns = [f"{prefix}{col_name}" for col_name in df_exploded.columns]

    return df.drop(columns=[col_name]).merge(
        df_exploded,
        how="left",
        left_index=True,
        right_index=True,
        suffixes=(None, suffix),
    )



import warnings

from stratosphere.utils.environment import is_colab, is_pyodide

with warnings.catch_warnings():
    warnings.simplefilter("ignore")

    if is_colab():
        from tqdm import tqdm
    else:
        from tqdm.auto import tqdm

    if is_pyodide():
        tqdm.monitor_interval = 0


def progress(*args, **kwargs):
    return tqdm(*args, **kwargs)



import contextlib
import cProfile
import io
import pstats


@contextlib.contextmanager
def profiled(print_callers: bool = False):
    """
    Profile the execution of a function, copied from
    https://docs.sqlalchemy.org/en/14/faq/performance.html

    Args:
        print_callers (bool, optional): IF true, print who's calling waht.
        Defaults to False.
    """
    pr = cProfile.Profile()
    pr.enable()
    yield
    pr.disable()
    s = io.StringIO()
    ps = pstats.Stats(pr, stream=s).sort_stats("cumulative")
    ps.print_stats()
    if print_callers:
        ps.print_callers()
    print(s.getvalue())  # noqa



import copy
import logging
from typing import Callable, List, Union

import pandas as pd
from stratosphere import config
from stratosphere.experiment import Experiment
from stratosphere.store.db.database import Database
from stratosphere.utils import log
from stratosphere.utils.enums import IfExists, enforce_enum
from stratosphere.utils.environment import get_version_stratosphere


class Stratosphere:
    """Instantiate the Stratosphere handler, it can then be sued
    to create, query, manage experiments. You can instantiate one or more,
    they do work together nicely without interfering.
    """

    def __init__(self, url: str = config.db.default_url, echo: bool = False, ask_password=False):
        """Create a new Stratosphere handler.

        Args:
            url (str, optional): Database URL. Defaults to config.db.default_url.
            echo (bool, optional): Enable SQLAlchemy logging. Defaults to False.
        """

        log.init_logging(config.log.level)
        self.db = Database(url, echo=echo, ask_password=ask_password)

    @log.default_exception_handler
    def define(
        self,
        name: str = None,
        funcs: Union[Callable, List[Callable]] = None,
        kwargs: dict = None,
        fields: dict = None,
        properties: dict = None,
        param_grid: List = None,
    ) -> Experiment:
        """Define a new experiment.

        Args:
            name (str, optional): Name of the experiment (must be unique). Defaults to None.
            funcs (Union[Callable, List[Callable]], optional): Function(s) we want to apply to
            the input parameters. Defaults to None. kwargs (dict, optional): Fixed arguments to
            the functions. Defaults to None. attributes (dict, optional): Fixed experiment properties
            to be tracked. Defaults to None. parameter_grid (List, optional): Variable parameters
            to pass to the functions. Defaults to None.

        Returns:
            Experiment: The newly defined experiment, reaady to be executed.
        """

        return Experiment(
            db=self.db,
            name=name,
            funcs=funcs,
            kwargs=kwargs,
            fields=fields,
            properties=properties,
            param_grid=param_grid,
        )

    @log.default_exception_handler
    def ls(self, include_properties=False) -> pd.DataFrame:
        """List the tracked experiments.

        Returns:
            pd.DataFrame: Pandas dataframe.
        """
        return Experiment.ls(self.db, include_properties=include_properties)

    @log.default_exception_handler
    def load(self, name: str = None, pickle=False) -> Experiment:
        """Load an experiment from the database and return it.

        Args:
            name (str, optional): The name of the experiment. Defaults to None.

        Returns:
            Experiment: The loaded experiment.
        """
        return Experiment.load(self.db, name, pickle=pickle)

    @log.default_exception_handler
    def persist(self, experiment: Experiment, if_exists: IfExists = "fail") -> Experiment:
        """Persist the experiment to the database (not necessarily
        the one used to track it initially).

        Args:
            experiment (Experiment): Experiment object to be persisted.
            if_exists (IfExists, optional): Either "replace" of "fail" in case
            the experiment name is already present. Defaults to "fail".

        Returns:
            Experiment: The persisted experiment.
        """

        # Enforce if_exists value
        if_exists = enforce_enum(if_exists, IfExists)

        # We make a copy of the experiment, and change the database reference.
        # We can then use all its methods, including persist(...).
        experiment = copy.deepcopy(experiment)
        experiment.db = self.db
        experiment.persist(if_exists=if_exists)
        return experiment

    def query(self, query: str, verbose: bool = False) -> pd.DataFrame:
        """Query the experiment's table

        Args:
            query (str): SQL query. "{ee}" occurrences are substituted with the experiments table name.
            verbose (bool, optional): If true, print the resulting query. Defaults to False.

        Returns:
            pd.DataFrame: Result of the query.
        """

        return self.db.pandas(query.format(ee=config.db.experiments_table), verbose=verbose)

    def version(self):
        """Log Stratosphere version"""
        logging.info(f"Stratosphere v{get_version_stratosphere()}")



import logging
import random
from functools import partial
from typing import Callable, List, Union

import joblib
import pandas as pd
from sqlalchemy.orm import load_only
from stratosphere import config
from stratosphere.job import Job
from stratosphere.run import Run, Runs, normalize_funcs
from stratosphere.store import serialization
from stratosphere.store.db import models
from stratosphere.store.db.database import Database, next_ulid, pandas_query, sanitize_table_name
from stratosphere.store.frames import reorder_columns
from stratosphere.utils import log
from stratosphere.utils.enums import IfExists, enforce_enum
from stratosphere.utils.environment import get_environment


class PickleNotFoundException(Exception):
    def __init__(self):
        self.message = "The pickle blob was not found."
        super().__init__(self.message)


class RunException(Exception):
    """Raised if there's a failure during the execution of functions on runs."""

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


class ExperimentAlreadyExists(Exception):
    """Exception to handle the attempt to add one more experiment with the same name."""

    def __init__(self, name):
        self.message = f"Experiment '{name}' already existing, use if_exists=\"replace\" to overwrite it."
        super().__init__(self.message)


class ExperimentNotFoundException(Exception):
    """Exception raised if the experiment is not found."""

    def __init__(self, name):
        self.message = f"Experiment '{name}' not found. List with .ls() the available experiments."
        super().__init__(self.message)


class Experiment:
    """This class represents a new experiment.

    Raises:
        ExperimentNotFoundException: Raised if we try to use a non-existing experiment.
        ExperimentAlreadyExists: Raised if the experiment name is already present.

    Returns:
        Experiment: A defined experiment, ready to be executed, tracked, stored, queried.
    """

    # model_cls is the SQLAlchemy model mapped to this class.
    model_cls = models.Experiment

    def __init__(
        self,
        db: Database = None,
        id_experiment: str = None,
        name: str = None,
        funcs: Union[Callable, List[Callable]] = None,
        kwargs: dict = None,
        fields: dict = None,
        properties: dict = None,
        param_grid: List = None,
        runs: List[Run] = None,
    ):
        """Create a new experiment.

        Args:
            db (Database, optional): Database to link to. Defaults to None.
            id_experiment (str, optional): ID of the experiment to be used. Defaults to None.
            name (str, optional): Name of the experiment. Defaults to None.
            funcs (Union[Callable, List[Callable]], optional): One or more functions to call. Defaults to None.
            kwargs (dict, optional): Fixed parameters passed to the functions. Defaults to None.
            fields (dict, optional): Fixed experiment fields to be tracked. Defaults to None.
            properties (dict, optional): Experiment properties, used to hold meta-attributes. Defaults to None.
            param_grid (List, optional): Grid of parameters, used to generate the run parameters.
            runs (List[Run], optional): List of runs belonging to the experiment. Defaults to None.
        """

        self.db = db
        self.id_experiment = next_ulid() if id_experiment is None else id_experiment
        self.name = name
        self.funcs = normalize_funcs(funcs)
        self.kwargs = {} if kwargs is None else kwargs
        self.fields = {} if fields is None else fields

        # Populate properties
        self.properties = {} if properties is None else properties
        if "env" not in self.properties:
            # If the experiment is loaded, then "env" is already populated and we retain it.
            self.properties["env"] = get_environment()

        # Populate param_grid
        self.param_grid = [{}] if (param_grid is None or param_grid == []) else param_grid
        if isinstance(self.param_grid, dict):
            self.param_grid = [self.param_grid]

        self.runs = Runs(runs)

    def __reduce__(self):
        """
        Make the expriment pickable, avoiding unpickable objects in SQLAlchemy.

        Returns:
            Experiment: An unpickled Experiment object.
        """
        return self.__class__, (
            None,
            self.id_experiment,
            self.name,
            self.funcs,
            self.kwargs,
            self.fields,
            self.properties,
            self.param_grid,
            self.runs,
        )

    def table_name(self) -> str:
        """Return the table name of the experiment

        Returns:
            str: Table name
        """

        return sanitize_table_name(f"{config.db.experiment_table_prefix}{self.name}")

    @log.default_exception_handler
    def query(self, query: str = "SELECT * FROM {table_name}", verbose: bool = False) -> pd.DataFrame:
        """Query the experiment's table

        Args:
            query (str, optional): SQL query. {e} is replaced with the table name of the experiment.,
            {id} with the experiment ID, and {ee} with the experiments table. Defaults to the complete record.
            verbose (bool, optional): If true, print the resulting query. Defaults to False.

        Returns:
            pd.DataFrame: Result of the query.
        """

        df = self.db.pandas(
            query.format(
                id=f"'{self.id_experiment}'",
                table_name=self.table_name(),
                experiments_table=config.db.experiments_table,
                name=f"'{self.name}'",
            ),
            verbose=verbose,
        )

        return df

    def load_runs(self, run_columns):
        # Retrieve all columns
        df = self.query(f"select * from {self.table_name()}")

        # Take care of deserialization
        for col_name in run_columns["serialized"]:
            df[col_name] = df[col_name].map(serialization.deserialize)

        # Reconstruct fields
        def series_to_run(row: pd.Series):
            fields = row[run_columns["serialized"] + run_columns["non_serialized"]].to_dict()
            run = Run(id_run=row["id_run"], fields=fields)
            return run

        self.runs = Runs(df.apply(lambda row: series_to_run(row), axis=1).values)

    @classmethod
    def load_pickle(cls, db: Database, name: str):
        logging.info(f"Loading experiment '{name}' (pickle)")

        with db.session() as session:
            record = (
                session.query(cls.model_cls)
                .options(load_only(Experiment.model_cls.pickle))
                .filter_by(name=name)
                .first()
            )

            if record is None:
                raise ExperimentNotFoundException(name)
            # Unpicke the serialized object.

            if record.pickle is None:
                raise PickleNotFoundException()

            experiment = serialization.pickle_loads(record.pickle)
            # Set the db of the experiment. Pickled db instances are never re-instantiated
            # completely, this let us reuse the existing database instance. See the Database
            # class for more comments on this.
            #
            # Important: this let us use SQLite memory databases, as only one instance is
            # used on all experiments.
            experiment.db = db
            return experiment

    @classmethod
    def load(cls, db: Database, name: str, pickle=False):
        """Load an experiment from database.

        Args:
            db (Database): Reference to a database.
            name (str): Name of the experiment.

        Raises:
            ExperimentNotFoundException: Raised if the experiment is not found.

        Returns:
            Experiment: Loaded experiment.
        """

        if pickle:
            return cls.load_pickle(db, name)

        logging.info(f"Loading experiment '{name}' (json-serialized)")

        # Create a new session
        columns = [
            Experiment.model_cls.id_experiment,
            Experiment.model_cls.id_experiment,
            Experiment.model_cls.name,
            Experiment.model_cls.run_count,
            Experiment.model_cls.run_columns,
            Experiment.model_cls.fields,
            Experiment.model_cls.properties,
            Experiment.model_cls.id_experiment,
        ]

        with db.session() as session:
            record = session.query(cls.model_cls).options(load_only(*columns)).filter_by(name=name).first()

            # If the record is not found, raise an error.
            if record is None:
                raise ExperimentNotFoundException(name)

            # Load the experiment from the record in "experiments" table.
            experiment = Experiment(
                db=db,
                id_experiment=record.id_experiment,
                name=record.name,
                fields=serialization.deserialize(record.fields),
                properties=serialization.deserialize(record.properties),
            )

            if record.run_count > 0:
                experiment.load_runs(run_columns=serialization.deserialize(record.run_columns))

            return experiment

    @log.default_exception_handler
    def execute(
        self,
        funcs: Union[Callable, List[Callable]] = None,
        backend=joblib.parallel.DEFAULT_BACKEND,
        n_jobs=-1,
    ):
        """Execute the experiment, then retuning it.

        Args:
            funcs (Union[Callable, List[Callable]], optional): Function(s) that will override the defined ones.
            If the value is "all", the re-evaluation of all defined functions is forced. Defaults to None.
            backend (_type_, optional): Joblib baackend. Defaults to joblib.parallel.DEFAULT_BACKEND.
            n_jobs (int, optional): Parallelization degree, defined as in Joblib. Defaults to -1.

        Returns:
            Run: the same experiment, ready for chained operations.
        """

        # Randomize the order of the runs to execute.
        random.shuffle(self.param_grid)

        if not self.runs:
            # First execution: create runs and execute them.
            # If more functions are passed, we append them for execution.
            tasks = [
                partial(
                    lambda run: run.execute(funcs=funcs),
                    Run(kwargs=self.kwargs, params=params, funcs=self.funcs),
                )
                for params in self.param_grid
            ]
        else:
            tasks = [
                partial(
                    lambda run: run.execute(funcs=funcs),
                    run,
                )
                for run in self.runs.values()
            ]

        if len(tasks) > 0:
            self.runs = Runs(Job(tasks, n_jobs=n_jobs, backend=backend).execute())

        # Check for exceptions, and report them if any.
        for run in self.runs.values():
            if run.exception is not None:
                raise RunException(run.exception)

        # If the execution is successful, we update the tracked environment.
        self.properties["env"] = get_environment()

        return self

    def record(
        self,
        run_columns=None,
        store_pickle=config.serialization.store_pickle,
        enable_compression=config.serialization.enable_compression,
    ) -> models.Experiment:
        """Build an SQLAlchemy ORM object from the existing Experiment object.

        Returns:
            models.Experiment: SQLAlchemy ORM object.
        """

        return self.model_cls(
            id_experiment=self.id_experiment,
            name=self.name,
            fields=serialization.serialize(self.fields, enable_compression=enable_compression),
            properties=serialization.serialize(self.properties, enable_compression=enable_compression),
            pickle=serialization.pickle_dumps(self) if store_pickle else None,
            run_count=len(self.runs),
            run_columns=serialization.serialize(run_columns, enable_compression=enable_compression),
        )

    def info(self):
        """Print some stats about the experiment.

        Returns:
            Experiment: self
        """
        logging.info("Experiment info")
        logging.info(f"  id...............: {self.id_experiment}")
        logging.info(f"  name.............: '{self.name}'")
        logging.info(f"  funcs............: {[func.__name__ for func in self.funcs]}")
        logging.info(f"  kwargs...........: {list(self.kwargs.keys())}")

        logging.info(f"  param_grid...: {list(self.param_grid[0].keys())} ({len(self.param_grid)})")

        logging.info(f"  fields...........: {list(self.fields.keys())}")
        logging.info(f"  size (Mb)........: {self.size('mb')} (compressed)")
        logging.info(f"  table name.......: {self.table_name()}")

        return self

    @log.default_exception_handler
    def persist(
        self,
        if_exists: IfExists = IfExists["fail"],
        store_pickle=config.serialization.store_pickle,
        enable_compression=config.serialization.enable_compression,
    ):
        """Persist the experiment.

        Args:
            if_exists (IfExists, optional): Either "replace" or "fail". Defaults to "fail".

        Raises:
            ExperimentReadOnlyException: Experiment loaded in SQL read-only mode.

        Returns:
            _type_: self.
        """

        logging.info(f"Persisting experiment (table name: {self.table_name()})")

        if_exists = enforce_enum(if_exists, IfExists)

        # Delete record and table of experiment, honoring if_exists.
        self.delete(if_exists)

        if self.runs:
            # The experiment has runs, let's find out the columns definition and persist it with the experiment.
            df_runs = self.runs.df(max_level=0)
            df_runs["id_experiment"] = self.id_experiment
            df_runs, run_columns = serialization.serialize_df(
                df_runs, ignore_columns=["id_run", "id_experiment"], enable_compression=enable_compression
            )
            df_runs = reorder_columns(df_runs, ["id_experiment", "id_run"])
        else:
            run_columns = None

        # Insert row in "experiments" table.
        with self.db.session() as session:
            session.add(
                self.record(run_columns=run_columns, store_pickle=store_pickle, enable_compression=enable_compression)
            )
            session.commit()

        if not self.runs:
            # This experiment has no runs, so we're done.
            return self

        dtype = {
            **{"id_run": models.uuid_type},
            **{"id_experiment": models.uuid_type},
            **{col_name: models.LargeBinary for col_name in run_columns["serialized"]},
        }

        # Create the experiment table.
        self.db.pandas_to_sql(df_runs, self.table_name(), if_exists.name, dtype=dtype)

        return self

    @log.default_exception_handler
    def delete(self, if_exists: IfExists = IfExists["fail"]) -> None:
        """Delete the expeirment.

        Args:
            if_exists (IfExists, optional): Either "replace" or "fail". In case
            the experiment exists and the value is "replace", nothing happens.
            If the experiment exists and the value is "fail", it will raise an
            exception. It is designed to work correctly together with insertions.
            Defaults to "fail".

        Raises:
            ExperimentAlreadyExists: Raised if the experiment exists.
        """

        if_exists = enforce_enum(if_exists, IfExists)

        # Create a new session
        with self.db.session() as session:
            # If we find the record ...
            if session.query(self.model_cls).filter_by(name=self.name).count() > 0:
                # And we are fine deleting it, proceed.
                if if_exists == IfExists["replace"]:
                    # We filter on matching experiment names, as these are the ones that might result
                    # in conflicts, since they must be unique.
                    session.query(Experiment.model_cls).filter(Experiment.model_cls.name == self.name).delete()
                else:
                    # Otherwise, raise an exception.
                    raise ExperimentAlreadyExists(self.name)
            session.commit()

        # We also need to drop the experiment table.
        self.db.drop_table(self.table_name())

    def df(self, max_level=None) -> pd.DataFrame:
        """Returns a Pandas dataframe representing the experiment, including
        parameters and fields. The processing does not depend on
        database queries, but the returned dataframe matches the contents
        of the experiment table (which is generated using this method).

        In presence of overlapping column names from the fixed fields
        of the experiment and the run fields, the fields use
        "_run" as suffix.

        Args:
            include_experiment (bool, optional): If False, do not include the
            fixed experiment fields (it will be a constant column).
            Defaults to True.

        Raises:
            ExperimentReadOnlyException: Experiment loaded in SQL read-only mode.

        Returns:
            pd.DataFrame: Pandas dataframe representing the tracked properties.
        """

        df_experiment = pd.json_normalize(
            {
                **self.fields,
                **{"id_experiment": self.id_experiment, "name": self.name},
            },
            max_level=max_level,
        )

        return reorder_columns(df_experiment, ["id_experiment", "name"])

    def size(self, unit: str = "b") -> int:
        """Return the size of the pickled version of the expeirment.

        Args:
            unit (str, optional): Unit of measure: "b" (Bytes), "kb" (KiloBytes), "mb" (MegaBytes). Defaults to "b".

        Returns:
            int: Size of the experiment once pickled.
        """
        return serialization.pickle_size(self, unit=unit)

    @classmethod
    def ls(cls, db: Database, include_properties=False) -> pd.DataFrame:
        """List experiments in a database.

        Args:
            db (Database): Database instance.

        Returns:
            pd.DataFrame: List of experiments.
        """

        # Construct list of columns to retrieve
        if include_properties:
            columns = [Experiment.model_cls.id_experiment, Experiment.model_cls.name, Experiment.model_cls.properties]
        else:
            columns = [Experiment.model_cls.id_experiment, Experiment.model_cls.name]

        with db.session() as session:
            df_experiments = pandas_query(
                session.query(Experiment.model_cls).options(load_only(*columns)),
                session,
            )

            df_experiments["table_name"] = df_experiments["name"].apply(
                lambda name: sanitize_table_name(f"{config.db.experiment_table_prefix}{name}")
            )

            if include_properties:
                df_experiments["properties"] = df_experiments["properties"].map(serialization.deserialize)
                df_experiments = serialization.explode_json_column(df_experiments, col_name="properties")

            return df_experiments
